@@ -26,17 +26,24 @@ if PY3:
 else:
     import Queue
 
-from pymongo import common
-from pymongo import periodic_executor
+from pymongo import (common,
+                     helpers,
+                     periodic_executor)
 from pymongo.pool import PoolOptions
 from pymongo.topology_description import (updated_topology_description,
                                           _updated_topology_description_srv_polling,
                                           TopologyDescription,
                                           SRV_POLLING_TOPOLOGIES, TOPOLOGY_TYPE)
-from pymongo.errors import ServerSelectionTimeoutError, ConfigurationError
+from pymongo.errors import (ConnectionFailure,
+                            ConfigurationError,
+                            NetworkTimeout,
+                            NotMasterError,
+                            OperationFailure,
+                            ServerSelectionTimeoutError)
 from pymongo.monitor import SrvMonitor
 from pymongo.monotonic import time as _time
 from pymongo.server import Server
+from pymongo.server_description import ServerDescription
 from pymongo.server_selectors import (any_server_selector,
                                       arbiter_server_selector,
                                       secondary_server_selector,
@@ -154,8 +161,12 @@ class Topology(object):
                 warnings.warn(
                     "MongoClient opened before fork. Create MongoClient only "
                     "after forking. See PyMongo's documentation for details: "
-                    "http://api.mongodb.org/python/current/faq.html#"
+                    "https://pymongo.readthedocs.io/en/stable/faq.html#"
                     "is-pymongo-fork-safe")
+                with self._lock:
+                    # Reset the session pool to avoid duplicate sessions in
+                    # the child process.
+                    self._session_pool.reset()
 
         with self._lock:
             self._ensure_opened()
@@ -202,7 +213,8 @@ class Topology(object):
             # No suitable servers.
             if timeout == 0 or now > end_time:
                 raise ServerSelectionTimeoutError(
-                    self._error_message(selector))
+                    "%s, Timeout: %ss, Topology Description: %r" %
+                    (self._error_message(selector), timeout, self.description))
 
             self._ensure_opened()
             self._request_check_all()
@@ -253,18 +265,23 @@ class Topology(object):
                                   server_selection_timeout,
                                   address)
 
-    def _process_change(self, server_description):
+    def _process_change(self, server_description, reset_pool=False):
         """Process a new ServerDescription on an opened topology.
 
         Hold the lock when calling this.
         """
         td_old = self._description
-        if self._publish_server:
-            old_server_description = td_old._server_descriptions[
-                server_description.address]
+        sd_old = td_old._server_descriptions[server_description.address]
+        if _is_stale_server_description(sd_old, server_description):
+            # This is a stale isMaster response. Ignore it.
+            return
+
+        suppress_event = ((self._publish_server or self._publish_tp)
+                          and sd_old == server_description)
+        if self._publish_server and not suppress_event:
             self._events.put((
                 self._listeners.publish_server_description_changed,
-                (old_server_description, server_description,
+                (sd_old, server_description,
                  server_description.address, self._topology_id)))
 
         self._description = updated_topology_description(
@@ -273,7 +290,7 @@ class Topology(object):
         self._update_servers()
         self._receive_cluster_time_no_lock(server_description.cluster_time)
 
-        if self._publish_tp:
+        if self._publish_tp and not suppress_event:
             self._events.put((
                 self._listeners.publish_topology_description_changed,
                 (td_old, self._description, self._topology_id)))
@@ -286,10 +303,16 @@ class Topology(object):
                                   SRV_POLLING_TOPOLOGIES):
             self._srv_monitor.close()
 
+        # Clear the pool from a failed heartbeat.
+        if reset_pool:
+            server = self._servers.get(server_description.address)
+            if server:
+                server.pool.reset()
+
         # Wake waiters in select_servers().
         self._condition.notify_all()
 
-    def on_change(self, server_description):
+    def on_change(self, server_description, reset_pool=False):
         """Process a new ServerDescription after an ismaster call completes."""
         # We do no I/O holding the lock.
         with self._lock:
@@ -303,7 +326,7 @@ class Topology(object):
             # that didn't include this server.
             if (self._opened and
                     self._description.has_server(server_description.address)):
-                self._process_change(server_description)
+                self._process_change(server_description, reset_pool)
 
     def _process_srv_update(self, seedlist):
         """Process a new seedlist on an opened topology.
@@ -397,37 +420,36 @@ class Topology(object):
             self._request_check_all()
             self._condition.wait(wait_time)
 
-    def reset_pool(self, address):
+    def handle_getlasterror(self, address, error_msg):
+        """Clear our pool for a server, mark it Unknown, and check it soon."""
+        error = NotMasterError(error_msg, {'code': 10107, 'errmsg': error_msg})
         with self._lock:
             server = self._servers.get(address)
             if server:
-                server.pool.reset()
+                self._process_change(
+                    ServerDescription(address, error=error), True)
+                server.request_check()
 
-    def reset_server(self, address):
-        """Clear our pool for a server and mark it Unknown.
+    def data_bearing_servers(self):
+        """Return a list of all data-bearing servers.
 
-        Do *not* request an immediate check.
+        This includes any server that might be selected for an operation.
         """
-        with self._lock:
-            self._reset_server(address, reset_pool=True)
+        if self._description.topology_type == TOPOLOGY_TYPE.Single:
+            return self._description.known_servers
+        return self._description.readable_servers
 
-    def reset_server_and_request_check(self, address):
-        """Clear our pool for a server, mark it Unknown, and check it soon."""
-        with self._lock:
-            self._reset_server(address, reset_pool=True)
-            self._request_check(address)
-
-    def mark_server_unknown_and_request_check(self, address):
-        """Mark a server Unknown, and check it soon."""
-        with self._lock:
-            self._reset_server(address, reset_pool=False)
-            self._request_check(address)
-
-    def update_pool(self):
+    def update_pool(self, all_credentials):
         # Remove any stale sockets and add new sockets if pool is too small.
+        servers = []
         with self._lock:
-            for server in self._servers.values():
-                server._pool.remove_stale_sockets()
+            # Only update pools for data-bearing servers.
+            for sd in self.data_bearing_servers():
+                server = self._servers[sd.address]
+                servers.append((server, server.pool.generation))
+
+        for server, generation in servers:
+            server.pool.remove_stale_sockets(generation, all_credentials)
 
     def close(self):
         """Clear pools and terminate monitors. Topology reopens on demand."""
@@ -529,29 +551,83 @@ class Topology(object):
         for server in itervalues(self._servers):
             server.open()
 
-    def _reset_server(self, address, reset_pool):
-        """Mark a server Unknown and optionally reset it's pool.
-
-        Hold the lock when calling this. Does *not* request an immediate check.
-        """
+    def _is_stale_error(self, address, err_ctx):
         server = self._servers.get(address)
+        if server is None:
+            # Another thread removed this server from the topology.
+            return True
 
-        # "server" is None if another thread removed it from the topology.
-        if server:
-            if reset_pool:
+        if err_ctx.sock_generation != server._pool.generation:
+            # This is an outdated error from a previous pool version.
+            return True
+
+        # topologyVersion check, ignore error when cur_tv >= error_tv:
+        cur_tv = server.description.topology_version
+        error = err_ctx.error
+        error_tv = None
+        if error and hasattr(error, 'details'):
+            if isinstance(error.details, dict):
+                error_tv = error.details.get('topologyVersion')
+
+        return _is_stale_error_topology_version(cur_tv, error_tv)
+
+    def _handle_error(self, address, err_ctx):
+        if self._is_stale_error(address, err_ctx):
+            return
+
+        server = self._servers[address]
+        error = err_ctx.error
+        exc_type = type(error)
+        if (issubclass(exc_type, NetworkTimeout) and
+                err_ctx.completed_handshake):
+            # The socket has been closed. Don't reset the server.
+            # Server Discovery And Monitoring Spec: "When an application
+            # operation fails because of any network error besides a socket
+            # timeout...."
+            return
+        elif issubclass(exc_type, NotMasterError):
+            # As per the SDAM spec if:
+            #   - the server sees a "not master" error, and
+            #   - the server is not shutting down, and
+            #   - the server version is >= 4.2, then
+            # we keep the existing connection pool, but mark the server type
+            # as Unknown and request an immediate check of the server.
+            # Otherwise, we clear the connection pool, mark the server as
+            # Unknown and request an immediate check of the server.
+            err_code = error.details.get('code', -1)
+            is_shutting_down = err_code in helpers._SHUTDOWN_CODES
+            # Mark server Unknown, clear the pool, and request check.
+            self._process_change(ServerDescription(address, error=error))
+            if is_shutting_down or (err_ctx.max_wire_version <= 7):
+                # Clear the pool.
+                server.reset()
+            server.request_check()
+        elif issubclass(exc_type, ConnectionFailure):
+            # "Client MUST replace the server's description with type Unknown
+            # ... MUST NOT request an immediate check of the server."
+            self._process_change(ServerDescription(address, error=error))
+            # Clear the pool.
+            server.reset()
+            # "When a client marks a server Unknown from `Network error when
+            # reading or writing`_, clients MUST cancel the isMaster check on
+            # that server and close the current monitoring connection."
+            server._monitor.cancel_check()
+        elif issubclass(exc_type, OperationFailure):
+            # Do not request an immediate check since the server is likely
+            # shutting down.
+            if error.code in helpers._NOT_MASTER_CODES:
+                self._process_change(ServerDescription(address, error=error))
+                # Clear the pool.
                 server.reset()
 
-            # Mark this server Unknown.
-            self._description = self._description.reset_server(address)
-            self._update_servers()
+    def handle_error(self, address, err_ctx):
+        """Handle an application error.
 
-    def _request_check(self, address):
-        """Wake one monitor. Hold the lock when calling this."""
-        server = self._servers.get(address)
-
-        # "server" is None if another thread removed it from the topology.
-        if server:
-            server.request_check()
+        May reset the server to Unknown, clear the pool, and request an
+        immediate check depending on the error and the context.
+        """
+        with self._lock:
+            self._handle_error(address, err_ctx)
 
     def _request_check_all(self):
         """Wake all monitors. Hold the lock when calling this."""
@@ -676,3 +752,38 @@ class Topology(object):
             else:
                 return ','.join(str(server.error) for server in servers
                                 if server.error)
+
+    def __repr__(self):
+        msg = ''
+        if not self._opened:
+            msg = 'CLOSED '
+        return '<%s %s%r>' % (self.__class__.__name__, msg, self._description)
+
+
+class _ErrorContext(object):
+    """An error with context for SDAM error handling."""
+    def __init__(self, error, max_wire_version, sock_generation,
+                 completed_handshake):
+        self.error = error
+        self.max_wire_version = max_wire_version
+        self.sock_generation = sock_generation
+        self.completed_handshake = completed_handshake
+
+
+def _is_stale_error_topology_version(current_tv, error_tv):
+    """Return True if the error's topologyVersion is <= current."""
+    if current_tv is None or error_tv is None:
+        return False
+    if current_tv['processId'] != error_tv['processId']:
+        return False
+    return current_tv['counter'] >= error_tv['counter']
+
+
+def _is_stale_server_description(current_sd, new_sd):
+    """Return True if the new topologyVersion is < current."""
+    current_tv, new_tv = current_sd.topology_version, new_sd.topology_version
+    if current_tv is None or new_tv is None:
+        return False
+    if current_tv['processId'] != new_tv['processId']:
+        return False
+    return current_tv['counter'] > new_tv['counter']
